import argparse
import shlex
from dataclasses import dataclass
from typing import Any

import pytest
from simple_parsing import ArgumentParser, field, DashVariant
from simple_parsing.wrappers.field_wrapper import FieldWrapper

from .testutils import *
from typing import List


def test_custom_args():
    @dataclass
    class Foo(TestSetup):
        output_dir: str = field(
            default="/out", alias=["-o", "--out"], choices=["/out", "/bob"]
        )

    foo = Foo.setup("--output_dir /bob")
    assert foo.output_dir == "/bob"

    with raises(SystemExit):
        foo = Foo.setup("-o /cat")
        assert foo.output_dir == "/cat"

    foo = Foo.setup("--out /bob")
    assert foo.output_dir == "/bob"


def test_custom_action_args():
    value = 0

    class CustomAction(argparse.Action):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, parser, namespace, values, dest):
            nonlocal value
            value += 1
            setattr(namespace, self.dest, values)

    @dataclass
    class Foo(TestSetup):
        output_dir: str = field(type=str, nargs="?", action=CustomAction)

    foo = Foo.setup("")
    assert value == 0

    foo = Foo.setup("--output_dir")
    assert value == 1

    foo = Foo.setup("--output_dir blablabob")
    assert value == 2
    assert foo.output_dir == "blablabob"


def test_custom_nargs_int():
    """Shows that you can use 'nargs' with the field() function."""

    @dataclass
    class Foo(TestSetup):
        output_dir: str = field(type=str, nargs=2)

    with raises_expected_n_args(2):
        foo = Foo.setup("--output_dir")

    with raises_expected_n_args(2):
        foo = Foo.setup("--output_dir hey")

    foo = Foo.setup("--output_dir john bob")
    assert foo.output_dir == ["john", "bob"]


def test_custom_nargs_plus():
    @dataclass
    class Foo(TestSetup):
        some_int: int = field(type=int, default=-1, nargs="+")

    with raises_missing_required_arg():
        foo = Foo.setup("")

    with exits_and_writes_to_stderr(match="expected at least one argument"):
        foo = Foo.setup("--some_int")

    foo = Foo.setup("--some_int 123")
    assert foo.some_int == [123]

    foo = Foo.setup("--some_int 123 456")
    assert foo.some_int == [123, 456]


def test_custom_nargs_star():
    @dataclass
    class Foo(TestSetup):
        some_int: int = field(type=int, nargs="*")

    foo = Foo.setup("")
    assert foo.some_int == None

    foo = Foo.setup("--some_int")
    assert foo.some_int == []

    foo = Foo.setup("--some_int 123")
    assert foo.some_int == [123]

    foo = Foo.setup("--some_int 123 456")
    assert foo.some_int == [123, 456]


def test_custom_nargs_question_mark():
    @dataclass
    class Foo(TestSetup):
        some_int: int = field(type=int, default=-1, nargs="?")

    foo = Foo.setup("")
    assert foo.some_int == -1

    foo = Foo.setup("--some_int")
    assert foo.some_int == None

    foo = Foo.setup("--some_int 123")
    assert foo.some_int == 123

    with raises_unrecognized_args("456"):
        foo = Foo.setup("--some_int 123 456")


@dataclass
class Foo:
    flag: bool = field(alias=["-f", "-flag"], action="store_true")
    # wether or not to store some value.
    no_cache: bool = field(action="store_false")


def test_store_true_action(parser: TestParser[Foo]):
    parser.add_arguments(Foo, "foo")
    foo = parser("--flag")
    assert foo.flag == True

    foo = parser("")
    assert foo.flag == False

    foo = parser("-f")
    assert foo.flag == True

    foo = parser("-flag")
    assert foo.flag == True


def test_store_false_action():
    parser = ArgumentParser(add_option_string_dash_variants=True)
    parser.add_arguments(Foo, "foo")

    args = parser.parse_args("--no-cache".split())
    foo: Foo = args.foo
    assert foo.no_cache == False

    args = parser.parse_args("".split())
    foo: Foo = args.foo
    assert foo.no_cache == True
 
def test_only_dashes():
    @dataclass
    class SomeClass(TestSetup):
        """lol"""
        my_var: int

    assert_help_output_equals(
        SomeClass.get_help_text(add_option_string_dash_variants=DashVariant.DASH),
        """
    usage: pytest [-h] --my-var int

    optional arguments:
    -h, --help            show this help message and exit

    test_only_dashes.<locals>.SomeClass ['some_class']:
    lol

      --my-var int
    """,
    )
    sc = SomeClass.setup('--my-var 2', add_option_string_dash_variants=DashVariant.DASH)
    assert sc.my_var == 2

def test_list_of_choices():
    @dataclass
    class Foo(TestSetup):
        """Some class Foo"""

        # A sequence of tasks.
        task_sequence: List[str] = field(choices=["train", "test", "ood"])

    foo = Foo.setup("--task_sequence train train ood")
    assert foo.task_sequence == ["train", "train", "ood"]

    with exits_and_writes_to_stderr(match="invalid choice:"):
        Foo.setup("--task_sequence train bob test")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--foo", action="store_const", const=42)
    args = parser.parse_args()
    print(args)



import argparse
from argparse import ONE_OR_MORE, OPTIONAL, PARSER, REMAINDER, ZERO_OR_MORE
from logging import getLogger
from typing import Type, Callable
from argparse import Action

from .utils import (
    get_type_arguments,
    get_type_name,
    is_optional,
    is_tuple,
    is_tuple_or_list,
    is_union,
)
from .wrappers.field_metavar import get_metavar

logger = getLogger(__name__)


class SimpleHelpFormatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.MetavarTypeHelpFormatter,
    argparse.RawDescriptionHelpFormatter,
):
    """Little shorthand for using some useful HelpFormatters from argparse.

    This class inherits from argparse's `ArgumentDefaultHelpFormatter`,
    `MetavarTypeHelpFormatter` and `RawDescriptionHelpFormatter` classes.

    This produces the following resulting actions:
    - adds a "(default: xyz)" for each argument with a default
    - uses the name of the argument type as the metavar. For example, gives
      "-n int" instead of "-n N" in the usage and description of the arguments.
    - Conserves the formatting of the class and argument docstrings, if given.
    """

    def _format_args(self, action: Action, default_metavar: str):
        _get_metavar = self._metavar_formatter(action, default_metavar)
        action_type = action.type

        metavar = action.metavar or get_metavar(action_type)
        if metavar and not action.choices:
            result = metavar
        elif action.nargs is None:
            result = "%s" % _get_metavar(1)
        elif action.nargs == OPTIONAL:
            result = "[%s]" % _get_metavar(1)
        elif action.nargs == ZERO_OR_MORE:
            result = "[%s [%s ...]]" % _get_metavar(2)
        elif action.nargs == ONE_OR_MORE:
            result = "%s [%s ...]" % _get_metavar(2)
        elif action.nargs == REMAINDER:
            result = "..."
        elif action.nargs == PARSER:
            result = "%s ..." % _get_metavar(1)
        else:
            formats = ["%s" for _ in range(action.nargs)]
            result = " ".join(formats) % _get_metavar(action.nargs)

        logger.debug(
            f"action type: {action_type}, Result: {result}, nargs: {action.nargs}, default metavar: {default_metavar}"
        )
        return result

    def _get_default_metavar_for_optional(self, action: argparse.Action):
        try:
            return super()._get_default_metavar_for_optional(action)
        except BaseException as e:
            logger.debug(f"Getting metavar for action with dest {action.dest}.")
            metavar = self._get_metavar_for_action(action)
            logger.debug(f"Result metavar: {metavar}")
            return metavar

    def _get_default_metavar_for_positional(self, action: argparse.Action):
        try:
            return super()._get_default_metavar_for_positional(action)
        except BaseException as e:
            logger.debug(f"Getting metavar for action with dest {action.dest}.")
            metavar = self._get_metavar_for_action(action)
            logger.debug(f"Result metavar: {metavar}")
            return metavar

    def _get_metavar_for_action(self, action: argparse.Action) -> str:
        return self._get_metavar_for_type(action.type)

    def _get_metavar_for_type(self, t: Type) -> str:
        return get_metavar(t) or str(t)


Formatter = SimpleHelpFormatter



import argparse
import shlex
import string
import sys
from contextlib import contextmanager, suppress
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)

import pytest

import simple_parsing
from simple_parsing import (
    ArgumentParser,
    ConflictResolution,
    InconsistentArgumentError,
    ParsingError,
    SimpleHelpFormatter,
    DashVariant,
)
from simple_parsing.utils import camel_case
from simple_parsing.wrappers import DataclassWrapper

xfail = pytest.mark.xfail
parametrize = pytest.mark.parametrize


def xfail_param(*args, reason: str):
    if len(args) == 1 and isinstance(args[0], tuple):
        args = args[0]
    return pytest.param(*args, marks=pytest.mark.xfail(reason=reason))


Dataclass = TypeVar("Dataclass")


@contextmanager
def raises(exception=ParsingError, match=None, code: int = None):
    with pytest.raises(exception, match=match):
        yield


from io import StringIO
from contextlib import redirect_stderr


@contextmanager
def exits_and_writes_to_stderr(match: str = ""):
    s = StringIO()
    with redirect_stderr(s), raises(SystemExit):
        yield
    s.seek(0)
    err_string = s.read()
    if match:
        assert match in err_string, err_string
    else:
        assert err_string, err_string


@contextmanager
def raises_missing_required_arg():
    with exits_and_writes_to_stderr("the following arguments are required"):
        yield


@contextmanager
def raises_expected_n_args(n: int):
    with exits_and_writes_to_stderr(f"expected {n} arguments"):
        yield


@contextmanager
def raises_unrecognized_args(*args: str):
    with exits_and_writes_to_stderr(f"unrecognized arguments: " + " ".join(args or [])):
        yield


def assert_help_output_equals(actual: str, expected: str) -> bool:
    # Replace the start with `prog`, since the test runner might not always be
    # `pytest`, could also be __main__ when debugging with VSCode
    prog = sys.argv[0].split("/")[-1]
    if prog != "pytest":
        expected = expected.replace("usage: pytest", f"usage: {prog}")
    remove = string.punctuation + string.whitespace
    actual_str = "".join(actual.split())
    expected_str = "".join(expected.split())
    assert actual_str == expected_str, f"{actual_str} != {expected_str}"


T = TypeVar("T")


class TestParser(simple_parsing.ArgumentParser, Generic[T]):
    __test__ = False
    """ A parser subclass just used for testing.
    Makes the retrieval of the arguments a bit easier to read.
    """

    def __init__(self, *args, **kwargs):
        self._current_dest = None
        self._current_dataclass = None
        super().__init__(*args, **kwargs)

    def add_arguments(self, dataclass: Type, dest, prefix="", default=None):
        if self._current_dest == dest and self._current_dataclass == dataclass:
            return  # already added arguments for that dataclass.
        self._current_dest = dest
        self._current_dataclass = dataclass
        return super().add_arguments(dataclass, dest, prefix=prefix, default=default)

    def __call__(self, args: str) -> T:
        namespace = self.parse_args(shlex.split(args))
        value = getattr(namespace, self._current_dest)
        value = cast(T, value)
        return value


class TestSetup:
    @classmethod
    def setup(
        cls: Type[Dataclass],
        arguments: Optional[str] = "",
        dest: Optional[str] = None,
        default: Optional[Dataclass] = None,
        conflict_resolution_mode: ConflictResolution = ConflictResolution.AUTO,
        add_option_string_dash_variants: DashVariant = DashVariant.AUTO,
        parse_known_args: bool = False,
        attempt_to_reorder: bool = False,
    ) -> Dataclass:
        """Basic setup for a test.

        Keyword Arguments:
            arguments {Optional[str]} -- The arguments to pass to the parser (default: {""})
            dest {Optional[str]} -- the attribute where the argument should be stored. (default: {None})

        Returns:
            {cls}} -- the class's type.
        """
        parser = simple_parsing.ArgumentParser(
            conflict_resolution=conflict_resolution_mode,
            add_option_string_dash_variants=add_option_string_dash_variants,
        )
        if dest is None:
            dest = camel_case(cls.__name__)

        parser.add_arguments(cls, dest=dest, default=default)

        if arguments is None:
            if parse_known_args:
                args = parser.parse_known_args(attempt_to_reorder=attempt_to_reorder)
            else:
                args = parser.parse_args()
        else:
            splits = shlex.split(arguments)
            if parse_known_args:
                args, unknown_args = parser.parse_known_args(
                    splits, attempt_to_reorder=attempt_to_reorder
                )
            else:
                args = parser.parse_args(splits)
        assert hasattr(args, dest), f"attribute '{dest}' not found in args {args}"
        instance: Dataclass = getattr(args, dest)  # type: ignore
        delattr(args, dest)
        assert (
            args == argparse.Namespace()
        ), f"Namespace has leftover garbage values: {args}"

        instance = cast(Dataclass, instance)
        return instance

    @classmethod
    def setup_multiple(
        cls: Type[Dataclass], num_to_parse: int, arguments: Optional[str] = ""
    ) -> Tuple[Dataclass, ...]:
        conflict_resolution_mode: ConflictResolution = ConflictResolution.ALWAYS_MERGE

        parser = simple_parsing.ArgumentParser(
            conflict_resolution=conflict_resolution_mode
        )
        class_name = camel_case(cls.__name__)
        for i in range(num_to_parse):
            parser.add_arguments(cls, f"{class_name}_{i}")

        if arguments is None:
            args = parser.parse_args()
        else:
            splits = shlex.split(arguments)
            args = parser.parse_args(splits)

        return tuple(getattr(args, f"{class_name}_{i}") for i in range(num_to_parse))

    @classmethod
    def get_help_text(
        cls,
        multiple=False,
        conflict_resolution_mode: ConflictResolution = ConflictResolution.AUTO,
        add_option_string_dash_variants = DashVariant.AUTO,
    ) -> str:
        import contextlib
        from io import StringIO

        f = StringIO()
        with contextlib.suppress(SystemExit), contextlib.redirect_stdout(f):
            _ = cls.setup(
                "--help",
                conflict_resolution_mode=conflict_resolution_mode,
                add_option_string_dash_variants=add_option_string_dash_variants,
            )
        s = f.getvalue()
        return s


ListFormattingFunction = Callable[[List[Any]], str]
ListOfListsFormattingFunction = Callable[[List[List[Any]]], str]


def format_list_using_spaces(value_list: List[Any]) -> str:
    return " ".join(str(p) for p in value_list)


def format_list_using_brackets(value_list: List[Any]) -> str:
    return f"[{','.join(str(p) for p in value_list)}]"


def format_list_using_single_quotes(value_list: List[Any]) -> str:
    return f"'{format_list_using_spaces(value_list)}'"


def format_list_using_double_quotes(value_list: List[Any]) -> str:
    return f'"{format_list_using_spaces(value_list)}"'


def format_lists_using_brackets(list_of_lists: List[List[Any]]) -> str:
    return " ".join(
        format_list_using_brackets(value_list) for value_list in list_of_lists
    )


def format_lists_using_double_quotes(list_of_lists: List[List[Any]]) -> str:
    return " ".join(
        format_list_using_double_quotes(value_list) for value_list in list_of_lists
    )


def format_lists_using_single_quotes(list_of_lists: List[List[Any]]) -> str:
    return " ".join(
        format_list_using_single_quotes(value_list) for value_list in list_of_lists
    )



from dataclasses import dataclass

import pytest

from .testutils import TestSetup


@dataclass
class Base(TestSetup):
    """Some extension of base-class `Base`"""

    a: int = 5
    f: bool = False


@dataclass
class Flags(TestSetup):
    a: bool  # an example required flag (defaults to False)
    b: bool = True  # optional flag 'b'.
    c: bool = False  # optional flag 'c'.


def test_bool_attributes_work():
    ext = Base.setup("--a 5 --f")
    assert ext.f == True

    ext = Base.setup("--a 5")
    assert ext.f == False

    true_strings = ["True", "true"]
    for s in true_strings:
        ext = Base.setup(f"--a 5 --f {s}")
        assert ext.f == True

    false_strings = ["False", "false"]
    for s in false_strings:
        ext = Base.setup(f"--a 5 --f {s}")
        assert ext.f == False


def test_bool_flags_work():
    flags = Flags.setup("--a true --b --c")
    assert flags.a == True
    assert flags.b == False
    assert flags.c == True



from dataclasses import dataclass, field
from typing import List, Optional

import pytest
from simple_parsing import ArgumentParser

from .testutils import TestSetup


@dataclass
class Config:
    seed: Optional[int] = None


def test_optional_seed():
    """Test that a value marked as Optional works fine.

    (Reproduces https://github.com/lebrice/SimpleParsing/issues/14#issue-562538623)
    """
    parser = ArgumentParser()
    parser.add_arguments(Config, dest="config")

    args = parser.parse_args("".split())
    config: Config = args.config
    assert config == Config()

    args = parser.parse_args("--seed 123".split())
    config: Config = args.config
    assert config == Config(123)


@dataclass
class Child:
    name: str = "Kevin"
    age: int = 12


@dataclass
class Dog:
    breed: str = "Saint Bernard"
    dog_years: int = 49


@dataclass
class Parent(TestSetup):
    child: Optional[Child] = None
    dog: Optional[Dog] = None


def test_optional_parameter_group():
    """Reproduces issue #28 :
    https://github.com/lebrice/SimpleParsing/issues/28#issue-663689719
    """
    parent: Parent = Parent.setup("--breed Shitzu")
    assert parent.dog == Dog(breed="Shitzu")
    assert parent.child == None

    parent: Parent = Parent.setup("--name Dylan")
    assert parent.dog == None
    assert parent.child == Child(name="Dylan")

    parent: Parent = Parent.setup("--name Dylan --dog_years 27")
    assert parent.dog == Dog(dog_years=27)
    assert parent.child == Child(name="Dylan")


@dataclass
class GrandParent(TestSetup):
    niece: Optional[Parent] = None
    nefew: Optional[Parent] = None


@pytest.mark.xfail(reason="TODO: Deeper nesting doesn't work atm!")
def test_deeply_nested_optional_parameter_groups():
    """Same as above test, but deeper hierarchy."""
    grandparent: GrandParent = GrandParent.setup()
    assert grandparent.niece == None
    assert grandparent.nefew == None

    grandparent: GrandParent = GrandParent.setup("--niece.child.name Bob")
    assert grandparent.niece == Parent(child=Child(name="Bob"))
    assert grandparent.nefew == None


def test_optional_int():
    @dataclass
    class Bob(TestSetup):
        num_workers: Optional[int] = 12

    assert Bob.setup("--num_workers") == Bob(num_workers=None)
    # assert Bob.setup("--num_workers None") == Bob(num_workers=None)
    assert Bob.setup("--num_workers 123") == Bob(num_workers=123)
    # assert Bob.setup("--num_workers none") == Bob(num_workers=None)


def test_optional_str():
    @dataclass
    class Bob(TestSetup):
        a: Optional[str] = "123"

    assert Bob.setup("--a") == Bob(a=None)
    # assert Bob.setup("--a None") == Bob(a="None")
    assert Bob.setup("--a 123") == Bob(a="123")
    # assert Bob.setup("--a none") == Bob(a="none")


@pytest.mark.xfail(reason=f"TODO: Rework the code to parse containers.")
def test_optional_list_of_ints():
    @dataclass
    class Bob(TestSetup):
        a: Optional[List[int]] = field(default_factory=list)

    # assert Bob.setup("--a") == Bob(a=None)
    assert Bob.setup("--a 1") == Bob(a=[1])
    assert Bob.setup("--a 1 2 3") == Bob(a=[1, 2, 3])
    assert Bob.setup("--a []") == Bob(a=[])
    assert Bob.setup("") == Bob(a=[])
    # assert Bob.setup("--a None") == Bob(a=None)
    # assert Bob.setup("--a none") == Bob(a=None)


def test_optional_without_default():
    @dataclass
    class Bob(TestSetup):
        a: Optional[int]

    assert Bob.setup("") == Bob(a=None)
    assert Bob.setup("--a") == Bob(a=None)
    # assert Bob.setup("--a None") == Bob(a=None)
    assert Bob.setup("--a 123") == Bob(a=123)
    # assert Bob.setup("--a none") == Bob(a=None)


# def test_simple_optional_argument(simple_attribute, silent):
#     some_type, passed_value, expected_value = simple_attribute
#     @dataclass
#     class SomeDataclass(TestSetup):
#         some_attribute: Optional[some_type] # type: ignore

#     actual = SomeDataclass.setup(f"--some_attribute {passed_value}")
#     assert actual.some_attribute == expected_value
#     assert isinstance(actual.some_attribute, some_type)



import argparse
import dataclasses
import shlex
from dataclasses import dataclass, field
from typing import *

import pytest

from .testutils import *
from simple_parsing import mutable_field
import simple_parsing.utils as utils
from simple_parsing.helpers import dict_field, set_field, list_field


@dataclass
class SomeDataclass:
    x: float = 123


@parametrize(
    "t",
    [
        Tuple[int, ...],
        Tuple[str],
        Tuple,
        tuple,
    ],
)
def test_is_tuple(t: Type):
    assert utils.is_tuple(t)
    assert not utils.is_list(t)


@parametrize(
    "t",
    [
        List[int],
        List[str],
        List,
        list,
        List[SomeDataclass],
    ],
)
def test_is_list(t: Type):
    assert utils.is_list(t)
    assert not utils.is_tuple(t)


@parametrize(
    "t",
    [
        List[SomeDataclass],
        Tuple[SomeDataclass],
    ],
)
def test_is_list_of_dataclasses(t: Type):
    assert utils.is_tuple_or_list_of_dataclasses(t)


@dataclass
class A:
    a: str = "bob"


@dataclass
class B:
    # # shared_list: List = [] # not allowed.
    # different_list: List = field(default_factory=list)
    shared: A = A()
    different: A = mutable_field(A, a="123")


def test_mutable_field():
    b1 = B()
    b2 = B()

    assert id(b1.shared) == id(
        b2.shared
    ), f"{b1.shared} should have the same id as {b2.shared}"
    assert id(b1.different) != id(b2.different), f"{b1.different} has the same id."


import enum


class Color(enum.Enum):
    RED = "RED"
    ORANGE = "ORANGE"
    BLUE = "BLUE"


class Temperature(enum.IntEnum):
    HOT = 1
    WARM = 0
    COLD = -1
    MONTREAL = -35


@parametrize(
    "t",
    [
        Color,
        Temperature,
    ],
)
def test_is_enum(t: Type):
    assert utils.is_enum(t)


def test_json_serializable(HyperParameters, tmpdir):
    hparams = HyperParameters()
    import os

    filename = "hparams.json"
    hparams.save_json(os.path.join(tmpdir, filename))
    hparams_ = HyperParameters.load_json(os.path.join(tmpdir, filename))
    assert hparams == hparams_


def test_list_field():
    @dataclass
    class A:
        a: List[str] = list_field("bob", "john", "bart")

    a1 = A()
    a2 = A()
    assert id(a1.a) != id(a2.a)


def test_set_field():
    @dataclass
    class A:
        a: Set[str] = set_field("bob", "john", "bart")

    a1 = A()
    a2 = A()
    assert id(a1.a) != id(a2.a)


def test_dict_field():
    default = {"bob": 0, "john": 1, "bart": 2}

    @dataclass
    class A:
        a: Dict[str, int] = dict_field(default)

    a1 = A()
    print(a1.a)
    assert a1.a == default
    a2 = A()
    assert id(a1.a) != id(a2.a)


def test_dict_field_with_keyword_args():
    default = {"bob": 0, "john": 1, "bart": 2}

    @dataclass
    class A(TestSetup):
        a: Dict[str, int] = dict_field(dict(bob=0, john=1, bart=2))

    a1 = A()
    a2 = A()
    assert a1.a == a2.a == default
    assert id(a1.a) != id(a2.a)


def test_dict_field_without_args():
    default = {}

    @dataclass
    class A(TestSetup):
        a: Dict[str, int] = dict_field()

    a1 = A()
    a2 = A()
    assert a1.a == a2.a == default
    assert id(a1.a) != id(a2.a)



import json
import sys
import warnings
from collections import OrderedDict
from dataclasses import MISSING, Field, dataclass, fields, is_dataclass
from itertools import chain
from logging import getLogger
from pathlib import Path
from typing import IO, Any, ClassVar, Dict, List, Optional, Tuple, Type, TypeVar, Union

from simple_parsing.utils import is_optional, get_args, get_forward_arg
from .decoding import decode_field, register_decoding_fn
from .encoding import SimpleJsonEncoder, encode

logger = getLogger(__name__)

Dataclass = TypeVar("Dataclass")
D = TypeVar("D", bound="Serializable")

try:
    import yaml

    def ordered_dict_constructor(loader: yaml.Loader, node: yaml.Node):
        # NOTE(ycho): `deep` has to be true for `construct_yaml_seq`.
        value = loader.construct_sequence(node, deep=True)
        return OrderedDict(*value)

    def ordered_dict_representer(
        dumper: yaml.Dumper, instance: OrderedDict
    ) -> yaml.Node:
        # NOTE(ycho): nested list for compatibility with PyYAML's representer
        node = dumper.represent_sequence("OrderedDict", [list(instance.items())])
        return node

    yaml.add_representer(OrderedDict, ordered_dict_representer)
    yaml.add_constructor("OrderedDict", ordered_dict_constructor)
    yaml.add_constructor(
        "tag:yaml.org,2002:python/object/apply:collections.OrderedDict",
        ordered_dict_constructor,
    )

except ImportError:
    pass


class SerializableMixin:
    """Makes a dataclass serializable to and from dictionaries.

    Supports JSON and YAML files for now.

    >>> from dataclasses import dataclass
    >>> from simple_parsing.helpers import Serializable
    >>> @dataclass
    ... class Config(Serializable):
    ...   a: int = 123
    ...   b: str = "456"
    ...
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.to_dict()
    {'a': 123, 'b': '456'}
    >>> config_ = Config.from_dict({"a": 123, "b": 456})
    >>> config_
    Config(a=123, b='456')
    >>> assert config == config_
    """

    subclasses: ClassVar[List[Type[D]]] = []
    decode_into_subclasses: ClassVar[bool] = False

    def __init_subclass__(
        cls, decode_into_subclasses: bool = None, add_variants: bool = True
    ):
        logger.debug(f"Registering a new Serializable subclass: {cls}")
        super().__init_subclass__()
        if decode_into_subclasses is None:
            # if decode_into_subclasses is None, we will use the value of the
            # parent class, if it is also a subclass of Serializable.
            # Skip the class itself as well as object.
            parents = cls.mro()[1:-1]
            logger.debug(f"parents: {parents}")

            for parent in parents:
                if parent in SerializableMixin.subclasses and parent is not SerializableMixin:
                    decode_into_subclasses = parent.decode_into_subclasses
                    logger.debug(
                        f"Parent class {parent} has decode_into_subclasses = {decode_into_subclasses}"
                    )
                    break

        cls.decode_into_subclasses = decode_into_subclasses or False
        if cls not in SerializableMixin.subclasses:
            SerializableMixin.subclasses.append(cls)

        encode.register(cls, cls.to_dict)
        register_decoding_fn(cls, cls.from_dict)

    def to_dict(self, dict_factory: Type[Dict] = dict, recurse: bool = True) -> Dict:
        """Serializes this dataclass to a dict.

        NOTE: This 'extends' the `asdict()` function from
        the `dataclasses` package, allowing us to not include some fields in the
        dict, or to perform some kind of custom encoding (for instance,
        detaching `Tensor` objects before serializing the dataclass to a dict).
        """
        d: Dict[str, Any] = dict_factory()
        for f in fields(self):
            name = f.name
            value = getattr(self, name)

            # Do not include in dict if some corresponding flag was set in metadata.
            include_in_dict = f.metadata.get("to_dict", True)
            if not include_in_dict:
                continue

            custom_encoding_fn = f.metadata.get("encoding_fn")
            if custom_encoding_fn:
                # Use a custom encoding function if there is one.
                d[name] = custom_encoding_fn(value)
                continue

            encoding_fn = encode
            if isinstance(value, SerializableMixin) and recurse:
                try:
                    encoded = value.to_dict(dict_factory=dict_factory, recurse=recurse)
                except TypeError:
                    encoded = value.to_dict()
                logger.debug(f"Encoded Serializable field {name}: {encoded}")
            else:
                try:
                    encoded = encoding_fn(value)
                except Exception as e:
                    logger.error(
                        f"Unable to encode value {value} of type {type(value)}! Leaving it as-is. (exception: {e})"
                    )
                    encoded = value
            d[name] = encoded
        return d

    @classmethod
    def from_dict(cls: Type[D], obj: Dict, drop_extra_fields: bool = None) -> D:
        """Parses an instance of `cls` from the given dict.

        NOTE: If the `decode_into_subclasses` class attribute is set to True (or
        if `decode_into_subclasses=True` was passed in the class definition),
        then if there are keys in the dict that aren't fields of the dataclass,
        this will decode the dict into an instance the first subclass of `cls`
        which has all required field names present in the dictionary.

        Passing `drop_extra_fields=None` (default) will use the class attribute
        described above.
        Passing `drop_extra_fields=True` will decode the dict into an instance
        of `cls` and drop the extra keys in the dict.
        Passing `drop_extra_fields=False` forces the above-mentioned behaviour.
        """
        if drop_extra_fields is None:
            drop_extra_fields = not cls.decode_into_subclasses
        return from_dict(cls, obj, drop_extra_fields=drop_extra_fields)

    def dump(self, fp: IO[str], dump_fn=json.dump, **kwargs) -> None:
        # Convert `self` into a dict.
        d = self.to_dict()
        # Serialize that dict to the file, using dump_fn.
        dump_fn(d, fp, **kwargs)

    def dump_json(self, fp: IO[str], dump_fn=json.dump, **kwargs) -> str:
        return self.dump(fp, dump_fn=dump_fn, **kwargs)

    def dump_yaml(self, fp: IO[str], dump_fn=None, **kwargs) -> str:
        import yaml

        if dump_fn is None:
            dump_fn = yaml.dump
        return self.dump(fp, dump_fn=dump_fn, **kwargs)

    def dumps(self, dump_fn=json.dumps, **kwargs) -> str:
        d = self.to_dict()
        return dump_fn(d, **kwargs)

    def dumps_json(self, dump_fn=json.dumps, **kwargs) -> str:
        kwargs.setdefault("cls", SimpleJsonEncoder)
        return self.dumps(dump_fn=dump_fn, **kwargs)

    def dumps_yaml(self, dump_fn=None, **kwargs) -> str:
        import yaml

        if dump_fn is None:
            dump_fn = yaml.dump
        return self.dumps(dump_fn=dump_fn, **kwargs)

    @classmethod
    def load(
        cls: Type[D],
        path: Union[Path, str, IO[str]],
        drop_extra_fields: bool = None,
        load_fn=None,
        **kwargs,
    ) -> D:
        """Loads an instance of `cls` from the given file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[Path, str, IO[str]]): Path or Path string or open file.
            drop_extra_fields (bool, optional): Wether to drop extra fields or
                to decode the dictionary into the first subclass with matching
                fields. Defaults to None, in which case we use the value of
                `cls.decode_into_subclasses`.
                For more info, see `cls.from_dict`.
            load_fn ([type], optional): Which loading function to use. Defaults
                to None, in which case we try to use the appropriate loading
                function depending on `path.suffix`:
                {
                    ".yml": yaml.safe_load,
                    ".yaml": yaml.safe_load,
                    ".json": json.load,
                    ".pth": torch.load,
                    ".pkl": pickle.load,
                }

        Raises:
            RuntimeError: If the extension of `path` is unsupported.

        Returns:
            D: An instance of `cls`.
        """
        if isinstance(path, str):
            path = Path(path)

        if load_fn is None and isinstance(path, Path):
            if path.name.endswith((".yml", ".yaml")):
                return cls.load_yaml(
                    path, drop_extra_fields=drop_extra_fields, **kwargs
                )
            elif path.name.endswith(".json"):
                return cls.load_json(
                    path, drop_extra_fields=drop_extra_fields, **kwargs
                )
            elif path.name.endswith(".pth"):
                import torch

                load_fn = torch.loads
            elif path.name.endswith(".npy"):
                import numpy as np

                load_fn = np.load
            elif path.name.endswith(".pkl"):
                import pickle

                load_fn = pickle.load
            warnings.warn(
                RuntimeWarning(
                    f"Not sure how to deserialize contents of {path} to a dict, as no "
                    f" load_fn was passed explicitly. Will try to use {load_fn} as the "
                    f"load function, based on the path name."
                )
            )

        if load_fn is None:
            raise RuntimeError(
                f"Unable to determine what function to use in order to load "
                f"path {path} into a dictionary, since no load_fn was passed, "
                f"and the path doesn't have an unfamiliar extension.."
            )

        if isinstance(path, Path):
            path = path.open()
        return cls._load(
            path, load_fn=load_fn, drop_extra_fields=drop_extra_fields, **kwargs
        )

    @classmethod
    def _load(
        cls: Type[D],
        fp: IO[str],
        drop_extra_fields: bool = None,
        load_fn=json.load,
        **kwargs,
    ) -> D:
        # Load a dict from the file.
        d = load_fn(fp, **kwargs)
        # Convert the dict into an instance of the class.
        return cls.from_dict(d, drop_extra_fields=drop_extra_fields)

    @classmethod
    def load_json(
        cls: Type[D],
        path: Union[str, Path],
        drop_extra_fields: bool = None,
        load_fn=json.load,
        **kwargs,
    ) -> D:
        """Loads an instance from the corresponding json-formatted file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[str, Path]): Path to a json-formatted file.
            load_fn ([type], optional): Loading function to use. Defaults to json.load.

        Returns:
            D: an instance of the dataclass.
        """
        return cls.load(
            path, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs
        )

    @classmethod
    def load_yaml(
        cls: Type[D],
        path: Union[str, Path],
        drop_extra_fields: bool = None,
        load_fn=None,
        **kwargs,
    ) -> D:
        """Loads an instance from the corresponding yaml-formatted file.

        Args:
            cls (Type[D]): A dataclass type to load.
            path (Union[str, Path]): Path to a yaml-formatted file.
            load_fn ([type], optional): Loading function to use. Defaults to
                None, in which case `yaml.safe_load` is used.

        Returns:
            D: an instance of the dataclass.
        """
        import yaml

        if load_fn is None:
            load_fn = yaml.safe_load
        return cls.load(
            path, load_fn=load_fn, drop_extra_fields=drop_extra_fields, **kwargs
        )

    def save(self, path: Union[str, Path], dump_fn=None, **kwargs) -> None:
        if not isinstance(path, Path):
            path = Path(path)

        if dump_fn is None and isinstance(path, Path):
            if path.name.endswith((".yml", ".yaml")):
                return self.save_yaml(path, **kwargs)
            elif path.name.endswith(".json"):
                return self.save_json(path, **kwargs)
            elif path.name.endswith(".pth"):
                import torch

                dump_fn = torch.save
            elif path.name.endswith(".npy"):
                import numpy as np

                dump_fn = np.save
            elif path.name.endswith(".pkl"):
                import pickle

                dump_fn = pickle.dump
            warnings.warn(
                RuntimeWarning(
                    f"Not 100% sure how to deserialize contents of {path} to a "
                    f"file as no dump_fn was passed explicitly. Will try to use "
                    f"{dump_fn} as the serialization function, based on the path "
                    f"suffix. ({path.suffix})"
                )
            )

        if dump_fn is None:
            raise RuntimeError(
                f"Unable to determine what function to use in order to dump "
                f"path {path} into a dictionary, since no dump_fn was passed, "
                f"and the path doesn't have an unfamiliar extension: "
                f"({path.suffix})"
            )
        self._save(path, dump_fn=dump_fn, **kwargs)

    def _save(self, path: Union[str, Path], dump_fn=json.dump, **kwargs) -> None:
        d = self.to_dict()
        logger.debug(f"saving to path {path}")
        with open(path, "w") as fp:
            dump_fn(d, fp, **kwargs)

    def save_yaml(self, path: Union[str, Path], dump_fn=None, **kwargs) -> None:
        import yaml

        if dump_fn is None:
            dump_fn = yaml.dump
        self.save(path, dump_fn=dump_fn, **kwargs)

    def save_json(self, path: Union[str, Path], dump_fn=json.dump, **kwargs) -> None:
        self.save(path, dump_fn=dump_fn, **kwargs)

    @classmethod
    def loads(
        cls: Type[D],
        s: str,
        drop_extra_fields: bool = None,
        load_fn=json.loads,
        **kwargs,
    ) -> D:
        d = load_fn(s, **kwargs)
        return cls.from_dict(d, drop_extra_fields=drop_extra_fields)

    @classmethod
    def loads_json(
        cls: Type[D],
        s: str,
        drop_extra_fields: bool = None,
        load_fn=json.loads,
        **kwargs,
    ) -> D:
        return cls.loads(
            s, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs
        )

    @classmethod
    def loads_yaml(
        cls: Type[D], s: str, drop_extra_fields: bool = None, load_fn=None, **kwargs
    ) -> D:
        import yaml

        if load_fn is None:
            load_fn = yaml.safe_load
        return cls.loads(s, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)


@dataclass
class Serializable(SerializableMixin):
    """Makes a dataclass serializable to and from dictionaries.

    Supports JSON and YAML files for now.

    >>> from dataclasses import dataclass
    >>> from simple_parsing.helpers import Serializable
    >>> @dataclass
    ... class Config(Serializable):
    ...   a: int = 123
    ...   b: str = "456"
    ...
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.to_dict()
    {'a': 123, 'b': '456'}
    >>> config_ = Config.from_dict({"a": 123, "b": 456})
    >>> config_
    Config(a=123, b='456')
    >>> assert config == config_
    """


@dataclass(frozen=True)
class FrozenSerializable(SerializableMixin):
    """Makes a (frozen) dataclass serializable to and from dictionaries.

    Supports JSON and YAML files for now.

    >>> from dataclasses import dataclass
    >>> from simple_parsing.helpers import Serializable
    >>> @dataclass
    ... class Config(Serializable):
    ...   a: int = 123
    ...   b: str = "456"
    ...
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.to_dict()
    {'a': 123, 'b': '456'}
    >>> config_ = Config.from_dict({"a": 123, "b": 456})
    >>> config_
    Config(a=123, b='456')
    >>> assert config == config_
    """


@dataclass
class SimpleSerializable(SerializableMixin, decode_into_subclasses=True):
    pass


S = TypeVar("S", bound=SerializableMixin)


def get_dataclass_types_from_forward_ref(
    forward_ref: Type, serializable_base_class: Type[S] = SerializableMixin
) -> List[Type[S]]:
    arg = get_forward_arg(forward_ref)
    potential_classes: List[Type] = []
    for serializable_class in serializable_base_class.subclasses:
        if serializable_class.__name__ == arg:
            potential_classes.append(serializable_class)
    return potential_classes


def from_dict(cls: Type[Dataclass], d: Dict[str, Any], drop_extra_fields: bool = None) -> Dataclass:
    """Parses an instance of the dataclass `cls` from the dict `d`.

    Args:
        cls (Type[Dataclass]): A `dataclass` type.
        d (Dict[str, Any]): A dictionary of `raw` values, obtained for example
            when deserializing a json file into an instance of class `cls`.
        drop_extra_fields (bool, optional): Wether or not to drop extra
            dictionary keys (dataclass fields) when encountered. There are three
            options:
            - True:
                The extra keys are dropped, and this function returns an
                instance of `cls`.
            - False:
                The extra keys (if any) are kept, and we search through the
                subclasses of `cls` for the first dataclass which has all the
                required fields.
            - None (default):
                `drop_extra_fields = not cls.decode_into_subclasses`.

    Raises:
        RuntimeError: If an error is encountered while instantiating the class.

    Returns:
        Dataclass: An instance of the dataclass `cls`.
    """
    if d is None:
        return None
    obj_dict: Dict[str, Any] = d.copy()

    init_args: Dict[str, Any] = {}
    non_init_args: Dict[str, Any] = {}

    if drop_extra_fields is None:
        drop_extra_fields = not getattr(cls, "decode_into_subclasses", False)
        logger.debug("drop_extra_fields is None. Using cls attribute.")

        if cls in {Serializable, FrozenSerializable, SerializableMixin}:
            # Passing `Serializable` means that we want to find the right
            # subclass depending on the keys.
            # We set the value to False when `Serializable` is passed, since
            # we use this mechanism when we don't know which dataclass to use.
            logger.debug("cls is `SerializableMixin`, drop_extra_fields = False.")
            drop_extra_fields = False

    logger.debug(f"from_dict for {cls}, drop extra fields: {drop_extra_fields}")
    for field in fields(cls) if is_dataclass(cls) else []:
        name = field.name
        if name not in obj_dict:
            if (
                field.metadata.get("to_dict", True)
                and field.default is MISSING
                and field.default_factory is MISSING
            ):
                logger.warning(
                    f"Couldn't find the field '{name}' in the dict with keys " f"{list(d.keys())}"
                )
            continue

        raw_value = obj_dict.pop(name)
        field_value = decode_field(field, raw_value)

        if field.init:
            init_args[name] = field_value
        else:
            non_init_args[name] = field_value

    extra_args = obj_dict

    # If there are arguments left over in the dict after taking all fields.
    if extra_args:
        if drop_extra_fields:
            logger.warning(f"Dropping extra args {extra_args}")
            extra_args.clear()

        elif issubclass(cls, (Serializable, FrozenSerializable, SerializableMixin)):
            # Use the first Serializable derived class that has all the required
            # fields.
            logger.debug(f"Missing field names: {extra_args.keys()}")

            # Find all the "registered" subclasses of `cls`. (from Serializable)
            derived_classes: List[Type[SerializableMixin]] = []
            for subclass in cls.subclasses:
                if issubclass(subclass, cls) and subclass is not cls:
                    derived_classes.append(subclass)
            logger.debug(f"All serializable derived classes of {cls} available: {derived_classes}")

            # All the arguments that the dataclass should be able to accept in
            # its 'init'.
            req_init_field_names = set(chain(extra_args, init_args))

            # Sort the derived classes by their number of init fields, so that
            # we choose the first one with all the required fields.
            derived_classes.sort(key=lambda dc: len(get_init_fields(dc)))

            for child_class in derived_classes:
                logger.debug(
                    f"child class: {child_class.__name__}, mro: {child_class.mro()}"
                )
                child_init_fields: Dict[str, Field] = get_init_fields(child_class)
                child_init_field_names = set(child_init_fields.keys())

                if child_init_field_names >= req_init_field_names:
                    # `child_class` is the first class with all required fields.
                    logger.debug(f"Using class {child_class} instead of {cls}")
                    return from_dict(child_class, d, drop_extra_fields=False)

    init_args.update(extra_args)
    try:
        instance = cls(**init_args)  # type: ignore
    except TypeError as e:
        # raise RuntimeError(f"Couldn't instantiate class {cls} using init args {init_args}.")
        raise RuntimeError(
            f"Couldn't instantiate class {cls} using init args {init_args.keys()}: {e}"
        )

    for name, value in non_init_args.items():
        logger.debug(f"Setting non-init field '{name}' on the instance.")
        setattr(instance, name, value)
    return instance


def get_init_fields(dataclass: Type) -> Dict[str, Field]:
    result: Dict[str, Field] = {}
    for field in fields(dataclass):
        if field.init:
            result[field.name] = field
    return result


def get_first_non_None_type(optional_type: Union[Type, Tuple[Type, ...]]) -> Optional[Type]:
    if not isinstance(optional_type, tuple):
        optional_type = get_args(optional_type)
    for arg in optional_type:
        if arg is not Union and arg is not type(None):
            logger.debug(f"arg: {arg} is not union? {arg is not Union}")
            logger.debug(f"arg is not type(None)? {arg is not type(None)}")
            return arg
    return None


def is_dataclass_or_optional_dataclass_type(t: Type) -> bool:
    """Returns wether `t` is a dataclass type or an Optional[<dataclass type>]."""
    return is_dataclass(t) or (
        is_optional(t) and is_dataclass(get_args(t)[0])
    )



from dataclasses import dataclass

from .testutils import *


def test_no_default_argument(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class")

    args = parser.parse_args(shlex.split(f"--a {passed_value}"))
    assert args == argparse.Namespace(some_class=SomeClass(a=expected_value))

    with raises_missing_required_arg():
        parser.parse_args("")


def test_default_dataclass_argument(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type

    parser = ArgumentParser()
    parser.add_arguments(
        SomeClass, dest="some_class", default=SomeClass(a=expected_value)
    )

    args = parser.parse_args("")
    assert args == argparse.Namespace(some_class=SomeClass(a=expected_value))


def test_default_dict_argument(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class", default={"a": expected_value})

    args = parser.parse_args("")
    assert args == argparse.Namespace(some_class=SomeClass(a=expected_value))


def test_default_dict_argument_override_cmdline(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class", default={"a": 0})

    args = parser.parse_args(shlex.split(f"--a {passed_value}"))
    assert args == argparse.Namespace(some_class=SomeClass(a=expected_value))


def test_partial_default_dict_argument(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type
        b: int

    parser = ArgumentParser()
    parser.add_arguments(SomeClass, dest="some_class", default={"a": expected_value})

    args = parser.parse_args(shlex.split("--b 0"))
    assert args == argparse.Namespace(some_class=SomeClass(a=expected_value, b=0))
    with raises_missing_required_arg():
        parser.parse_args(shlex.split(f"--a {passed_value}"))
    with raises_missing_required_arg():
        parser.parse_args("")



from dataclasses import dataclass
import simple_parsing
import textwrap
import pytest


@dataclass
class JBuildRelease:
    id: int
    url: str
    docker_image: str


def test_issue_46(assert_equals_stdout):
    parser = simple_parsing.ArgumentParser()
    parser.add_argument("--run_id", type=str)
    parser.add_arguments(JBuildRelease, dest="jbuild", prefix="jbuild")

    parser.print_help()

    assert_equals_stdout(
        textwrap.dedent(
            """\
        usage: pytest [-h] [--run_id str] --jbuildid int --jbuildurl str
                      --jbuilddocker_image str
        
        optional arguments:
          -h, --help            show this help message and exit
          --run_id str
        
        JBuildRelease ['jbuild']:
          JBuildRelease(id:int, url:str, docker_image:str)
        
          --jbuildid int
          --jbuildurl str
          --jbuilddocker_image str
        """
        )
    )
    from .testutils import raises_missing_required_arg

    with raises_missing_required_arg():
        args = parser.parse_args(
            "--id 123 --jbuild.id 456 --jbuild.url bob --jbuild.docker_image foo".split()
        )


def test_issue_46_solution2(assert_equals_stdout):
    # This (now) works:
    parser = simple_parsing.ArgumentParser(add_dest_to_option_strings=True)
    parser.add_argument("--run_id", type=str)
    parser.add_arguments(JBuildRelease, dest="jbuild", prefix="jbuild.")

    parser.print_help()
    assert_equals_stdout(
        textwrap.dedent(
            """\
        usage: pytest [-h] [--run_id str] --jbuild.id int --jbuild.url str
                      --jbuild.docker_image str

        optional arguments:
          -h, --help            show this help message and exit
          --run_id str

        JBuildRelease ['jbuild']:
          JBuildRelease(id:int, url:str, docker_image:str)

          --jbuild.id int
          --jbuild.url str
          --jbuild.docker_image str
        """
        )
    )


@pytest.mark.xfail(reason="TODO: Issue #49")
def test_conflict_with_regular_argparse_arg():
    # This _should_ work, but it doesn't, adding a new issue for this:
    # the problem: SimpleParsing doesn't yet detect
    # conflicts between arguments added the usual way with `add_argument` and those
    # added through `add_arguments`.
    parser = simple_parsing.ArgumentParser()
    parser.add_argument("--id", type=str)
    parser.add_arguments(JBuildRelease, dest="jbuild")
    args = parser.parse_args(
        "--id 123 --jbuild.id 456 --jbuild.url bob --jbuild.docker_image foo".split()
    )
    assert args.id == 123
    assert args.jbuild.id == 456


@pytest.mark.xfail(reason="TODO: Issue #49")
def test_workaround():
    from simple_parsing import mutable_field, ConflictResolution

    # This also doesn't work, since the prefix is only added to the 'offending'
    # argument, rather than to all the args in that group.
    @dataclass
    class Main:
        id: int
        jbuild: JBuildRelease

    parser = simple_parsing.ArgumentParser()
    parser.add_arguments(Main, "main")
    args = parser.parse_args(
        "--id 123 --jbuild.id 456 --jbuild.url bob --jbuild.docker_image foo".split()
    )
    args = args.main
    assert args.id == 123
    assert args.jbuild.id == 456



"""Simple, extendable mechanism for encoding pracitaclly anything to string.

Just register a new encoder for a given type like so:

from simple_parsing.helpers.serialization import encode
import numpy as np
@encode.register
def encode_ndarray(obj: np.ndarray) -> str:
    return obj.tostring()
"""
import copy
import json
from dataclasses import fields, is_dataclass
from functools import singledispatch
from logging import getLogger
from os import PathLike
from typing import Any, Dict, Hashable, List, Set, Tuple, Union
from collections.abc import Mapping
from argparse import Namespace


logger = getLogger(__name__)


class SimpleJsonEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        return encode(o)


"""
# NOTE: This code is commented because of static typing check error.
# The problem is incompatibility of mypy and singledispatch.
# See mypy issues for more info:
# https://github.com/python/mypy/issues/8356
# https://github.com/python/mypy/issues/2904
# https://github.com/python/mypy/issues/9112#issuecomment-725316936

class Dataclass(Protocol):
    # see dataclasses.is_dataclass implementation with _FIELDS
    __dataclass_fields__: Dict[str, Field[Any]]


T = TypeVar("T", bool, int, None, str)


@overload
def encode(obj: Dataclass) -> Dict[str, Any]: ...

@overload
def encode(obj: Union[List[Any], Set[Any], Tuple[Any, ...]]) -> List[Any]:
    ...

@overload
def encode(obj: Mapping[Any, Any]) -> Dict[Any, Any]: ...

@overload
def encode(obj: T) -> T: ...
"""


@singledispatch
def encode(obj: Any) -> Any:
    """Encodes an object into a json/yaml-compatible primitive type.

    This called to convert field attributes when calling `to_dict()` on a
    `DictSerializable` instance (including JsonSerializable and YamlSerializable).

    This is used as the 'default' keyword argument to `json.dumps` and
    `json.dump`, and is called when an object is encountered that `json` doesn't
    know how to serialize.

    To register a type as JsonSerializable, you can just register a custom
    serialization function. (There should be no need to do it for dataclasses,
    since that is supported by this function), use @encode.register
    (see the docs for singledispatch).
    """
    try:
        if is_dataclass(obj):
            # logger.debug(f"encoding object {obj} of class {type(obj)}")
            d: Dict[str, Any] = dict()
            for field in fields(obj):
                value = getattr(obj, field.name)
                try:
                    d[field.name] = encode(value)
                except TypeError as e:
                    logger.error(f"Unable to encode field {field.name}: {e}")
                    raise e
            return d
        else:
            # logger.debug(f"Deepcopying object {obj} of type {type(obj)}")
            return copy.deepcopy(obj)
    except Exception as e:
        logger.debug(f"Cannot encode object {obj}: {e}")
        raise e


@encode.register(list)
@encode.register(tuple)
# @encode.register(Sequence) # Would also encompass `str!`
@encode.register(set)
def encode_list(obj: Union[List[Any], Set[Any], Tuple[Any, ...]]) -> List[Any]:
    # TODO: Here we basically say "Encode all these types as lists before serializing"
    # That's ok for JSON, but YAML can serialize stuff directly though.
    # TODO: Also, with this, we also need to convert back to the right type when
    # deserializing, which is totally doable for the fields of dataclasses,
    # but maybe not for other stuff.
    return list(map(encode, obj))


@encode.register(Mapping)
def encode_dict(obj: Mapping) -> Dict[Any, Any]:
    constructor = type(obj)
    result = constructor()
    for k, v in obj.items():
        k_ = encode(k)
        v_ = encode(v)
        if isinstance(k_, Hashable):
            result[k_] = v_
        else:
            # If the encoded key isn't "Hashable", then we store it as a list of tuples
            if isinstance(result, dict):
                result = list(result.items())
            result.append((k_, v_))
    return result

    return type(obj)((encode(k), encode(v)) for k, v in obj.items())


@encode.register(PathLike)
def encode_path(obj: PathLike) -> str:
    return obj.__fspath__()


@encode.register(Namespace)
def encode_namespace(obj: Namespace) -> Any:
    return encode(vars(obj))



import shlex
from dataclasses import dataclass

import pytest

from simple_parsing import ArgumentParser
from simple_parsing.helpers import Serializable, list_field, choice
from .testutils import *


@dataclass
class Base(TestSetup):
    """Some extension of base-class `Base`"""

    a: int = 1


@dataclass
class ExtendedB(Base, TestSetup):
    b: int = 2


@dataclass
class ExtendedC(Base, TestSetup):
    c: int = 3


@dataclass
class Inheritance(TestSetup):
    ext_b: ExtendedB = ExtendedB()
    ext_c: ExtendedC = ExtendedC()


def test_simple_subclassing_no_args():
    extended = ExtendedB.setup()
    assert extended.a == 1
    assert extended.b == 2


def test_simple_subclassing_with_args():
    extended = ExtendedB.setup("--a 123 --b 56")
    assert extended.a == 123
    assert extended.b == 56


# @xfail(reason="TODO: make sure this is how people would want to use this feature.")
def test_subclasses_with_same_base_class_no_args():
    ext = Inheritance.setup()

    assert ext.ext_b.a == 1
    assert ext.ext_b.b == 2

    assert ext.ext_c.a == 1
    assert ext.ext_c.c == 3


def test_subclasses_with_same_base_class_with_args():
    ext = Inheritance.setup(
        "--ext_b.a 10 --b 20 --ext_c.a 30 --c 40",
        conflict_resolution_mode=ConflictResolution.AUTO,
    )
    assert ext.ext_b.a == 10
    assert ext.ext_b.b == 20

    assert ext.ext_c.a == 30
    assert ext.ext_c.c == 40


@xfail(
    reason=(
        "Merging is not working yet with triangle inheritance, since we wouldn't "
        "know how to assign which value to which attribute.."
    )
)
def test_subclasses_with_same_base_class_with_args_merge():
    ext = Inheritance.setup(
        "--a 10 30 --b 20 --c 40",
        conflict_resolution_mode=ConflictResolution.ALWAYS_MERGE,
    )

    assert ext.ext_b.a == 10
    assert ext.ext_b.b == 20

    assert ext.ext_c.a == 30
    assert ext.ext_c.c == 40


def test_weird_structure():
    """both is-a, and has-a at the same time, a very weird inheritance structure"""

    @dataclass
    class ConvBlock(Serializable):
        """A Block of Conv Layers."""

        n_layers: int = 4  # number of layers
        n_filters: List[int] = list_field(16, 32, 64, 64)  # filters per layer

    @dataclass
    class GeneratorHParams(ConvBlock):
        """Settings of the Generator model"""

        conv: ConvBlock = ConvBlock()
        optimizer: str = choice("ADAM", "RMSPROP", "SGD", default="ADAM")

    @dataclass
    class DiscriminatorHParams(ConvBlock):
        """Settings of the Discriminator model"""

        conv: ConvBlock = ConvBlock()
        optimizer: str = choice("ADAM", "RMSPROP", "SGD", default="ADAM")

    @dataclass
    class SomeWeirdClass(TestSetup):
        gen: GeneratorHParams
        disc: DiscriminatorHParams

    s = SomeWeirdClass.setup()
    assert s.gen.conv.n_layers == 4
    assert s.gen.n_layers == 4
    assert s.disc.conv.n_layers == 4
    assert s.disc.n_layers == 4



"""Functions that are to be used to parse a field.

Somewhat analogous to the 'parse' function in the
`helpers.serialization.parsing` package.
"""
import enum
import functools
from dataclasses import Field
from logging import getLogger
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

from simple_parsing.utils import (
    get_type_arguments,
    is_enum,
    is_homogeneous_tuple_type,
    is_list,
    is_tuple,
    is_union,
    str2bool,
    is_typevar,
    get_bound,
    is_forward_ref,
    get_forward_arg,
)

logger = getLogger(__name__)

T = TypeVar("T")
K = TypeVar("K")


# Dictionary mapping from types/type annotations to their parsing functions.
_parsing_fns: Dict[Type[T], Callable[[Any], T]] = {
    # the 'primitive' types are parsed using the type fn as a constructor.
    t: t
    for t in [str, float, int, bytes]
}
_parsing_fns[bool] = str2bool


def get_parsing_fn_for_field(field: Field) -> Callable[[Any], T]:
    """Gets the parsing function for the field `field`."""
    name = field.name
    field_type = field.type
    logger.debug(f"name = {name}, field_type = {field_type}")

    # If the user set a custom parsing function, we use it.
    custom_parsing_fn = field.metadata.get("type")
    if custom_parsing_fn is not None:
        return custom_parsing_fn

    parsing_fn = get_parsing_fn(field.type)
    return parsing_fn


def _register(t: Type, func: Callable) -> None:
    if t not in _parsing_fns:
        # logger.debug(f"Registering the type {t} with parsing function {func}")
        _parsing_fns[t] = func


def register_parsing_fn(some_type: Type[T], function: Callable[[Any], T]) -> None:
    """Register a parsing function for the type `some_type`."""
    _register(some_type, function)


# This doesn't work as well as it did for serialization, in large part due to how
# argparse uses the `type` function when parsing containers.
# TODO: Replace this with a simpler function that just returns the 'arg_options' dict to
# give for a given type annotation.
def get_parsing_fn(t: Type[T]) -> Callable[[Any], T]:
    """Gets a parsing function for the given type or type annotation.

    Args:
        t (Type[T]): A type or type annotation.

    Returns:
        Callable[[Any], T]: A function that will parse a value of the given type
            from the command-line when available, or a no-op function that
            will return the raw value, when a parsing fn cannot be found or
            constructed.
    """
    if t in _parsing_fns:
        logger.debug(f"The type {t} has a dedicated parsing function.")
        return _parsing_fns[t]

    elif t is Any:
        logger.debug(f"parsing an Any type: {t}")
        return no_op

    # TODO: Do we want to support parsing a Dict from command-line?
    # elif is_dict(t):
    #     logger.debug(f"parsing a Dict field: {t}")
    #     args = get_type_arguments(t)
    #     if len(args) != 2:
    #         args = (Any, Any)
    #     return parse_dict(*args)

    # TODO: This would require some sort of 'postprocessing' step to convert a
    # list to a Set or something like that.
    # elif is_set(t):
    #     logger.debug(f"parsing a Set field: {t}")
    #     args = get_type_arguments(t)
    #     if len(args) != 1:
    #         args = (Any,)
    #     return parse_set(args[0])

    elif is_tuple(t):
        logger.debug(f"parsing a Tuple field: {t}")
        args = get_type_arguments(t)
        if is_homogeneous_tuple_type(t):
            if not args:
                args = (str, ...)
            parsing_fn = get_parsing_fn(args[0])
        else:
            parsing_fn = parse_tuple(args)
            parsing_fn.__name__ = str(t)
        return parsing_fn

    elif is_list(t):
        logger.debug(f"parsing a List field: {t}")
        args = get_type_arguments(t)
        assert len(args) == 1
        return parse_list(args[0])

    elif is_union(t):
        logger.debug(f"parsing a Union field: {t}")
        args = get_type_arguments(t)
        return parse_union(*args)
    
    elif is_enum(t):
        logger.debug(f"Parsing an Enum field of type {t}")
        return parse_enum(t)
    # import typing_inspect as tpi
    # from .serializable import get_dataclass_type_from_forward_ref, Serializable

    if is_forward_ref(t):
        forward_arg = get_forward_arg(t)
        for t, fn in _parsing_fns.items():
            if getattr(t, "__name__", str(t)) == forward_arg:
                return fn

    if is_typevar(t):
        bound = get_bound(t)
        logger.debug(f"parsing a typevar: {t}, bound type is {bound}.")
        if bound is not None:
            return get_parsing_fn(bound)

    logger.debug(
        f"Couldn't find a parsing function for type {t}, will try "
        f"to use the type directly."
    )
    return t


def try_functions(*funcs: Callable[[Any], T]) -> Callable[[Any], Union[T, Any]]:
    """Tries to use the functions in succession, else returns the same value unchanged."""

    def _try_functions(val: Any) -> Union[T, Any]:
        logger.debug(f"Debugging the 'raw value' of {val}, will try functions {funcs}")
        e: Optional[Exception] = None
        for func in funcs:
            try:
                parsed = func(val)
                logger.debug(
                    f"Successfully used the function {func} to get a parsed value of {parsed}."
                )
                return parsed
            except Exception as ex:
                e = ex
        else:
            logger.error(
                f"Couldn't parse value {val}, returning it as-is. (exception: {e})"
            )
        return val

    return _try_functions


def parse_union(*types: Type[T]) -> Callable[[Any], Union[T, Any]]:
    types = list(types)
    optional = type(None) in types
    # Partion the Union into None and non-None types.
    while type(None) in types:
        types.remove(type(None))

    parsing_fns: List[Callable[[Any], T]] = [
        parse_optional(t) if optional else get_parsing_fn(t) for t in types
    ]
    # Try using each of the non-None types, in succession. Worst case, return the value.
    return try_functions(*parsing_fns)


def parse_optional(t: Type[T]) -> Callable[[Optional[Any]], Optional[T]]:
    parse = get_parsing_fn(t)

    def _parse_optional(val: Optional[Any]) -> Optional[T]:
        return val if val is None else parse(val)

    return _parse_optional


def parse_tuple(
    tuple_item_types: Tuple[Type[T], ...]
) -> Callable[[List[T]], Tuple[T, ...]]:
    """Makes a parsing function for creating tuples from the command-line args.

    Can handle tuples with different item types, for instance:
    - `Tuple[int, Foo, str, float, ...]`.

    Returns:
        Callable[[List[T]], Tuple[T, ...]]: A parsing function for creating tuples.
    """
    # Note, if there are more values than types in the tuple type, then the
    # last type is used.
    # TODO: support the Ellipsis?
    if not tuple_item_types:
        tuple_item_types = (Any, Ellipsis)

    calls_count: int = 0

    def _parse_tuple(val: Any) -> Tuple[T, ...]:
        nonlocal calls_count
        logger.debug(
            f"Parsing a Tuple with item types {tuple_item_types}, raw value is {val}."
        )
        parsing_fn_index = calls_count

        if Ellipsis in tuple_item_types:
            ellipsis_index = tuple_item_types.index(Ellipsis)
            logger.debug(f"Ellipsis is at index {ellipsis_index}")
            # If this function is being called for the 'Ellipsis' type argument
            # or higher, just use the last type argument before the ellipsis.
            # NOTE: AFAIK, using something like Tuple[t1, t2, ...] is impossible
            # and it can only be something like Tuple[t1, ...], meaning an
            # unknown number of arguments of type `t1`.
            if parsing_fn_index >= ellipsis_index:
                parsing_fn_index = ellipsis_index - 1

        item_type = tuple_item_types[parsing_fn_index]
        parsing_fn = get_parsing_fn(item_type)
        parsed_value = parsing_fn(val)

        calls_count += 1

        return parsed_value


    return _parse_tuple


def parse_list(list_item_type: Type[T]) -> T:
    return get_parsing_fn(list_item_type)


def no_op(v: T) -> T:
    """Parsing function that gives back the value as-is.

    Args:
        v ([Any]): Any value.

    Returns:
        [type]: The value unchanged.
    """
    return v


E = TypeVar("E", bound=enum.Enum)


def parse_enum(enum_type: Type[E]) -> Callable[[str], E]:
    """Returns a function to use to parse an enum of type `enum_type` from a string.

    Parameters
    ----------
    - enum_type : Type[enum.Enum]

        The type of enum to create a parsing function for.

    Returns
    -------
    Callable[[str], E]
        A function that parses an enum object of type `enum_type` from a string.
    """
    # Save the function, since the same type will always be parsed the same way. Also
    # makes testing easier.
    if enum_type in _parsing_fns:
        return _parsing_fns[enum_type]
    
    # NOTE: Use `functools.wraps` so that fn name is the enum, so the metavar shows up
    # just like the enum on the command-line, and not like
    # "(...).parse_enum.<locals>._parse_enum" or something.
    @functools.wraps(enum_type)
    def _parse_enum(v: str) -> E:
        return enum_type[v]
    _parsing_fns[enum_type] = _parse_enum
    return _parse_enum
 


from dataclasses import dataclass

import pytest

import simple_parsing
from simple_parsing.docstring import get_attribute_docstring
from simple_parsing import field
from typing import List
from test.testutils import TestSetup
from .testutils import TestSetup


@dataclass
class Base:
    """A simple base-class example"""

    a: int  # TODO: finetune this

    """docstring for attribute 'a'"""

    b: float = 5.0  # inline comment on attribute 'b'

    c: str = ""
    """Multi
    Line
    Docstring for 'c'
    """


@dataclass
class Extended(Base):
    """Some extension of base-class `Base`"""

    ## Comment above d)
    # its multiline, does it still work?
    d: int = 5
    """ docstring for 'd' in Extended. """

    # Comment above e, but with a line skipped

    e: float = -1  # *# comment on the side of e


def test_docstring_parsing_work_on_base():
    docstring = get_attribute_docstring(Base, "a")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "TODO: finetune this"
    assert docstring.docstring_below == "docstring for attribute 'a'"

    docstring = get_attribute_docstring(Base, "b")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "inline comment on attribute 'b'"
    assert docstring.docstring_below == ""

    docstring = get_attribute_docstring(Base, "c")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == ""
    assert docstring.docstring_below == "Multi\nLine\nDocstring for 'c'\n"


def test_docstring_parsing_works_on_extended():
    docstring = get_attribute_docstring(Extended, "a")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "TODO: finetune this"
    assert docstring.docstring_below == "docstring for attribute 'a'"

    docstring = get_attribute_docstring(Extended, "b")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == "inline comment on attribute 'b'"
    assert docstring.docstring_below == ""

    docstring = get_attribute_docstring(Extended, "c")
    assert docstring.comment_above == ""
    assert docstring.comment_inline == ""
    assert docstring.docstring_below == "Multi\nLine\nDocstring for 'c'\n"

    docstring = get_attribute_docstring(Extended, "d")
    assert (
        docstring.comment_above
        == "# Comment above d)\nits multiline, does it still work?"
    )
    assert docstring.comment_inline == ""
    assert docstring.docstring_below == "docstring for 'd' in Extended."

    docstring = get_attribute_docstring(Extended, "e")
    assert docstring.comment_above == "Comment above e, but with a line skipped"
    assert docstring.comment_inline == "*# comment on the side of e"
    assert docstring.docstring_below == ""


def test_docstring_works_with_field_function():
    @dataclass
    class Foo(TestSetup):
        """Some class Foo"""

        # A sequence of tasks.
        task_sequence: List[str] = field(choices=["train", "test", "ood"])  # side
        """Below"""

    docstring = get_attribute_docstring(Foo, "task_sequence")
    assert docstring.comment_above == "A sequence of tasks."
    assert docstring.comment_inline == "side"
    assert docstring.docstring_below == "Below"



import dataclasses
from functools import wraps
from logging import getLogger
from typing import Any, Dict, List, Optional, Type, TypeVar, Union, overload, Tuple

from simple_parsing.helpers.fields import choice as _choice
from simple_parsing.helpers.fields import field

from .priors import CategoricalPrior, LogUniformPrior, NormalPrior, Prior, UniformPrior

HP = TypeVar("HP", bound="HyperParameters")
logger = getLogger(__name__)
T = TypeVar("T")


class ValueOutsidePriorException(Exception):
    def __init__(self, value: Any, prior: Any, *args, **kwargs):
        self.value = value
        self.prior = prior
        super().__init__(*args, **kwargs)


@overload
def uniform(
    min: int,
    max: int,
    default: int = None,
    discrete: bool = True,
    strict: bool = False,
    shape: Union[int, Tuple[int, ...]] = None,
    **kwargs,
) -> int:
    pass


@overload
def uniform(
    min: float,
    max: float,
    default: float = None,
    strict: bool = False,
    shape: Union[int, Tuple[int, ...]] = None,
    **kwargs,
) -> float:
    pass


@overload
def uniform(
    min: float,
    max: float,
    default: float = None,
    discrete: bool = False,
    strict: bool = False,
    shape: Union[int, Tuple[int, ...]] = None,
    **kwargs,
) -> float:
    pass


def uniform(
    min: Union[int, float],
    max: Union[int, float],
    discrete: bool = None,
    default: Union[int, float, dataclasses._MISSING_TYPE] = dataclasses.MISSING,
    strict: bool = False,
    shape: Union[int, Tuple[int, ...]] = None,
    **kwargs,
) -> Union[int, float]:
    """Declares a Field with a Uniform prior.

    Parameters
    ----------
    min : Union[int, float]
        Minimum value.

    max : Union[int, float]
        Maximum value.

    discrete : bool, optional
        Wether this value is sampled discretely (an integer) or not. By default None, in
        which case the types of the `min`, `max` and `default` arguments
        will be used to infer wether the field should be discrete or not.

    default : Union[int, float], optional
        The default value to use. Not setting a value here makes this field a required
        argument of the dataclass. Setting a value can also be useful for the experiment
        version control feature of Orion. By default None.

    strict : bool, optional
        Wether the bounds should be strictly enforced. When set to True, attempting to
        create an object with this field, and passing a value outside the bounds will
        raise a ValueError.

    Returns
    -------
    Union[int, float]
        A `dataclasses.Field` object, with the prior stored in its metadata. The return
        type annotation is intentionally "wrong", so that the type checker doesnt raise
        errors when declaring fields on a dataclass, since their type annotations
        wouldn't match with the return type of this function.
    """
    # TODO: what about uniform over a "choice"?
    if "default_value" in kwargs:
        assert default in {
            None,
            dataclasses.MISSING,
        }, "can't pass both `default` and `default_value`"
        default = kwargs.pop("default_value")

    if discrete is None:
        # TODO: Set discrete = False by default, and then maybe set it to True later if
        # the annotation on the field is int.
        discrete = False
        if min == 0 and max == 1:
            discrete = False
        elif (isinstance(min, int) and isinstance(max, int)) and (
            default in {None, dataclasses.MISSING} or isinstance(default, int)
        ):
            # If given something like uniform(0, 100) or uniform(5,10,default=7) then
            # we can 'safely' assume that the discrete option should be used.
            discrete = True
    if shape and default not in {None, dataclasses.MISSING}:
        assert isinstance(shape, int), "only support int shapes for now."
        if isinstance(default, (int, float)):
            default = tuple(default for _ in range(shape))

    if discrete and default not in {None, dataclasses.MISSING}:
        if shape is None:
            default = round(default)
        else:
            assert isinstance(shape, int), "only support int shapes for now."
            default = tuple(round(default[i]) for i in range(shape))

    # TODO: Make sure this doesn't by accident make some fields behave as positional
    # fields
    default_v = None if default is dataclasses.MISSING else default

    prior = UniformPrior(
        min=min, max=max, discrete=discrete, default=default_v, shape=shape
    )

    # if default is None:
    #     default = dataclasses.MISSING
    #     default = (min + max) / 2

    return hparam(default=default, prior=prior, strict=strict, **kwargs)


@overload
def log_uniform(
    min: int,
    max: int,
    discrete: bool = True,
    shape: Union[int, Tuple[int, ...]] = None,
    **kwargs,
) -> int:
    pass


@overload
def log_uniform(
    min: float,
    max: float,
    discrete: bool = False,
    shape: Union[int, Tuple[int, ...]] = None,
    **kwargs,
) -> float:
    pass


def log_uniform(
    min: Union[int, float],
    max: Union[int, float],
    discrete: bool = False,
    default: Union[int, float, dataclasses._MISSING_TYPE] = dataclasses.MISSING,
    shape: Union[int, Tuple[int, ...]] = None,
    **kwargs,
) -> Union[int, float]:
    if "default_value" in kwargs:
        assert default in {
            None,
            dataclasses.MISSING,
        }, "can't pass both `default` and `default_value`"
        default = kwargs.pop("default_value")

    default_v = default
    if default is dataclasses.MISSING:
        default_v = None
    prior = LogUniformPrior(
        min=min, max=max, discrete=discrete, default=default_v, shape=shape
    )

    # TODO: Do we really want to set the default value when not passed?
    # if default in {None, dataclasses.MISSING}:
    #     log_min = math.log(min, prior.base)
    #     log_max = math.log(max, prior.base)
    #     default = math.pow(prior.base, (log_min + log_max) / 2)
    #     if discrete or (isinstance(min, int) and isinstance(max, int)):
    #         default = round(default)
    return hparam(
        default=default,
        prior=prior,
        **kwargs,
    )


loguniform = log_uniform


@wraps(_choice)
def categorical(
    *choices: T,
    default: Union[T, dataclasses._MISSING_TYPE] = dataclasses.MISSING,
    probabilities: Union[List[float], Dict[str, float]] = None,
    strict: bool = False,
    **kwargs: Any,
) -> T:
    """Marks a field as being a categorical hyper-parameter.

    This wraps the `choice` function from `simple_parsing`, making it possible to choose
    the value from the command-line.

    The probabilites for each value should be passed through this `probabilities`
    argument.

    Returns:
        T: the result of the usual `dataclasses.field()` function (a dataclass field).
    """
    if "default_value" in kwargs:
        assert default in {
            None,
            dataclasses.MISSING,
        }, "can't pass both `default` and `default_value`"
        default = kwargs.pop("default_value")

    metadata = kwargs.get("metadata", {})
    default_key = default

    options: List[Any]
    if len(choices) == 1 and isinstance(choices[0], dict):
        choice_dict = choices[0]
        if probabilities and not isinstance(probabilities, dict):
            raise RuntimeError(
                "Need to pass a dict of probabilites when passing a dict of choices."
            )
        # TODO: If we use keys here, then we have to add a step in __post_init__ of the
        # dataclass holding this field, so that it gets the corresponding value from the
        # dict.
        # IDEA: Adding some kind of 'hook' to be used by simple-parsing?

        def postprocess(value):
            if value in choice_dict:
                return choice_dict[value]
            return value

        metadata["postprocessing"] = postprocess
        if default not in {None, dataclasses.MISSING}:
            assert default in choice_dict.values()
            default_key = [k for k, v in choice_dict.items() if v == default][0]
        options = list(choice_dict.keys())
    else:
        options = list(choices)

    if isinstance(probabilities, dict):
        if abs(sum(probabilities.values()) - 1) > 1e-5:
            raise RuntimeError("Probabilities should sum to 1!")
        probs = []
        for option in options:
            probability = probabilities.get(option, 0.0)
            probs.append(probability)
        probabilities = probs

    default_v = default_key
    if default is dataclasses.MISSING:
        default_v = None

    prior = CategoricalPrior(
        choices=options,
        probabilities=probabilities,
        default_value=default_v,
    )
    metadata["prior"] = prior

    if strict:

        def postprocess(value):
            if value not in prior:
                raise ValueOutsidePriorException(value=value, prior=prior)
            return value

        if "postprocessing" not in metadata:
            metadata["postprocessing"] = postprocess
        else:
            # TODO: Compose both functions?
            existing_fn = metadata["postprocessing"]
            new_fn = postprocess
            metadata["postprocessing"] = lambda v: new_fn(existing_fn(v))

    kwargs["metadata"] = metadata
    return _choice(*choices, default=default, **kwargs)


def hparam(
    default: T,
    *args,
    prior: Union[Type[Prior[T]], Prior[T]] = None,
    strict: bool = False,
    **kwargs,
) -> T:
    metadata = kwargs.get("metadata", {})
    min: Optional[float] = kwargs.get("min", kwargs.get("min"))
    max: Optional[float] = kwargs.get("max", kwargs.get("max"))

    if prior is None:
        assert min is not None and max is not None
        # if min and max are passed but no Prior object, assume a Uniform prior.
        prior = UniformPrior(min=min, max=max)
        metadata.update(
            {
                "min": min,
                "max": max,
                "prior": prior,
            }
        )

    elif isinstance(prior, type) and issubclass(prior, (UniformPrior, LogUniformPrior)):
        # use the prior as a constructor.
        assert min is not None and max is not None
        prior = prior(min=min, max=max)

    elif isinstance(prior, Prior):
        metadata["prior"] = prior
        if isinstance(prior, (UniformPrior, LogUniformPrior)):
            metadata.update(
                dict(
                    min=prior.min,
                    max=prior.max,
                )
            )
        elif isinstance(prior, (NormalPrior)):
            metadata.update(
                dict(
                    mu=prior.mu,
                    sigma=prior.sigma,
                )
            )

    else:
        # TODO: maybe support an arbitrary callable?
        raise RuntimeError(
            "hparam should receive either: \n"
            "- `min` and `max` kwargs, \n"
            "- `min` and `max` kwargs and a type of Prior to use, \n"
            "- a `Prior` instance."
        )

    if strict:
        assert "postprocessing" not in metadata

        def postprocess(value):
            if value not in prior:
                raise ValueOutsidePriorException(value=value, prior=prior)
            return value

        metadata["postprocessing"] = postprocess

    kwargs["metadata"] = metadata
    assert "default" not in kwargs
    return field(
        default=default,
        *args,
        **kwargs,
    )



import json
import logging
import textwrap
from collections import OrderedDict
from dataclasses import dataclass, fields
from pathlib import Path
from test.conftest import silent
from test.testutils import *
from typing import Any, Dict, List, Mapping, Optional, Set, Tuple, Type, Union

import pytest
import yaml

from simple_parsing import field, mutable_field
from simple_parsing.helpers import (
    Serializable,
    YamlSerializable,
    dict_field,
    list_field,
)
from simple_parsing.helpers.serialization.decoding import (
    get_decoding_fn,
    register_decoding_fn,
)


def test_encode_something(simple_attribute):

    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass(Serializable):
        d: Dict[str, some_type] = dict_field()
        l: List[Tuple[some_type, some_type]] = list_field()
        t: Dict[str, Optional[some_type]] = dict_field()
        # w: Dict[str, Union[some_type, int, str, None, str, None]] = dict_field()

    b = SomeClass()
    b.d.update({"hey": expected_value})
    b.l.append((expected_value, expected_value))
    b.t.update({"hey": None, "hey2": expected_value})
    # b.w.update({
    #     "hey": None,
    #     "hey2": "heyo",
    #     "hey3": 1,
    #     "hey4": expected_value,
    # })
    assert SomeClass.loads(b.dumps()) == b


def test_typevar_decoding(simple_attribute):
    @dataclass
    class Item(Serializable, decode_into_subclasses=True):
        name: str = "chair"
        price: float = 399
        stock: int = 10

    @dataclass
    class DiscountedItem(Item):
        discount_factor: float = 0.5

    I = TypeVar("I", bound=Item)

    @dataclass
    class Container(Serializable, Generic[I]):
        items: List[I] = list_field()

    chair = Item()
    cheap_chair = DiscountedItem(name="Cheap chair")
    c = Container(items=[chair, cheap_chair])

    assert Container.loads(c.dumps()) == c

    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass(Serializable):
        d: Dict[str, some_type] = dict_field()
        l: List[Tuple[some_type, some_type]] = list_field()
        t: Dict[str, Optional[some_type]] = dict_field()
        # w: Dict[str, Union[some_type, int, str, None, str, None]] = dict_field()

    b = SomeClass()
    b.d.update({"hey": expected_value})
    b.l.append((expected_value, expected_value))
    b.t.update({"hey": None, "hey2": expected_value})
    # b.w.update({
    #     "hey": None,
    #     "hey2": "heyo",
    #     "hey3": 1,
    #     "hey4": expected_value,
    # })
    assert SomeClass.loads(b.dumps()) == b


def test_super_nesting():
    @dataclass
    class Complicated(Serializable):
        x: List[
            List[List[Dict[int, Tuple[int, float, str, List[float]]]]]
        ] = list_field()

    c = Complicated()
    c.x = [[[{0: (2, 1.23, "bob", [1.2, 1.3])}]]]
    assert Complicated.loads(c.dumps()) == c
    assert c.dumps() == '{"x": [[[{"0": [2, 1.23, "bob", [1.2, 1.3]]}]]]}'


@pytest.mark.parametrize(
    "some_type, encoded_value, expected_value",
    [
        # (Tuple[int, float], json.loads(json.dumps([1, 2])), (1, 2.0)),
        (
            List[Tuple[int, float]],
            json.loads(json.dumps([[1, 2], [3, 4]])),
            [(1, 2.0), (3, 4.0)],
        ),
    ],
)
def test_decode_tuple(some_type: Type, encoded_value: Any, expected_value: Any):
    decoding_function = get_decoding_fn(some_type)
    actual = decoding_function(encoded_value)
    assert actual == expected_value



""" Tests for compatibility with the postponed evaluation of annotations. """
from __future__ import annotations

from simple_parsing import ArgumentParser
import sys
import pytest
from .testutils import TestSetup
from dataclasses import fields

from dataclasses import dataclass, field
from .testutils import TestSetup


@dataclass
class Foo(TestSetup):
    a: int = 123
    
    b: str = "fooobar"
    c: tuple[int, float] = (123, 4.56)
    
    d: list[bool] = field(default_factory=list)


@dataclass
class Bar(TestSetup):
    barry: Foo = field(default_factory=Foo)
    joe: "Foo" = field(default_factory=lambda: Foo(b="rrrrr"))
    z: float = 123.456
    some_list: list[float] = field(default_factory=[1.0, 2.0].copy)


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Needs access to annotations from __future__")
def test_future_annotations():
    foo = Foo.setup()
    assert foo == Foo()

    foo = Foo.setup("--a 2 --b heyo --c 1 7.89")
    assert foo == Foo(a=2, b="heyo", c=(1, 7.89))


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Needs access to annotations from __future__")
def test_future_annotations_nested():
    bar = Bar.setup()
    assert bar == Bar()
    assert bar.barry == Foo()
    bar = Bar.setup("--barry.a 2 --barry.b heyo --barry.c 1 7.89")
    assert bar.barry == Foo(a=2, b="heyo", c=(1, 7.89))
    assert isinstance(bar.joe, Foo)
    



"""A test to make sure that all the example files work without crashing.
(Could be seen as a kind of integration test.)

"""
from collections import Counter
import contextlib
import glob
import importlib
import shlex
import sys
from io import StringIO
from typing import Callable, Optional
from pathlib import Path
import pytest

expected = ""


def xfail_param(*args, reason: str):
    return pytest.param(*args, marks=pytest.mark.xfail(reason=reason))


@pytest.fixture
def set_prog_name():
    argv = sys.argv.copy()
    del sys.argv[1:]

    def set_prog(prog_name: str, args: str):
        sys.argv[0] = prog_name
        sys.argv[1:] = shlex.split(args)

    yield set_prog
    sys.argv = argv


import textwrap


@pytest.fixture
def assert_equals_stdout(capsys):
    def strip(string):
        return "".join(string.split())

    import difflib

    def should_equal(expected: str, file_path: str):
        out = capsys.readouterr().out
        # assert strip(out) == strip(expected), file_path
        # assert out == expected, file_path
        out_lines = out.splitlines(keepends=False)
        out_lines = [line.strip() for line in out_lines if line and not line.isspace()]
        expected_lines = expected.splitlines(keepends=False)
        expected_lines = [
            line.strip() for line in expected_lines if line and not line.isspace()
        ]
        if sys.version_info[:2] >= (3, 9):
            assert Counter("".join(out_lines)) == Counter("".join(expected_lines))
        else:
            assert out_lines == expected_lines, file_path

    return should_equal


@pytest.mark.parametrize(
    "file_path, args",
    [
        ("examples/subparsers/subparsers_example.py", "train"),
    ],
)
def test_running_example_outputs_expected(
    file_path: str,
    args: str,
    set_prog_name: Callable[[str, Optional[str]], None],
    assert_equals_stdout: Callable[[str, str], None],
):
    script = file_path.split("/")[-1]
    # set_prog_name(script, args)
    file_path = Path(file_path).as_posix()
    module_name = file_path.replace("/", ".").replace(".py", "")
    try:
        # programmatically import the example script, which also runs it.
        # (Equivalent to executing "from <module_name> import expected")
        import runpy

        resulting_globals = runpy.run_module(
            module_name, init_globals=None, run_name="__main__", alter_sys=True
        )
        # module = __import__(module_name, globals(), locals(), ["expected"], 0)
        # resulting_globals = vars(module)
        # get the 'expected'
        # assert "expected" in resulting_globals
        if "expected" not in resulting_globals:
            pytest.xfail(reason="Example doesn't have an 'expected' global variable.")
        expected = resulting_globals["expected"]
        assert_equals_stdout(expected, file_path)

    except SystemExit as e:
        pytest.xfail(f"SystemExit in example {file_path}.")


@pytest.mark.parametrize(
    "file_path",
    [
        *[
            p
            for p in glob.glob("examples/**/*.py")
            if p
            not in {
                "examples/merging/multiple_lists_example.py",
                "examples/subparsers/subparsers_example.py",
                "examples/serialization/custom_types_example.py",
            }
        ],
    ],
)
def test_running_example_outputs_expected_without_arg(
    file_path: str,
    set_prog_name: Callable[[str, Optional[str]], None],
    assert_equals_stdout: Callable[[str, str], None],
):
    return test_running_example_outputs_expected(
        file_path, "", set_prog_name, assert_equals_stdout
    )



"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
from . import helpers, utils, wrappers
from .conflicts import ConflictResolution
from .help_formatter import SimpleHelpFormatter
from .helpers import (
    MutableField,
    Serializable,
    choice,
    field,
    flag,
    list_field,
    mutable_field,
    subparsers,
)
from .parsing import ArgumentParser, ParsingError, DashVariant
from .utils import InconsistentArgumentError

__all__ = [
    "helpers",
    "utils",
    "wrappers",
    "ConflictResolution",
    "DashVariant"
    "MutableField",
    "Serializable",
    "SimpleHelpFormatter",
    "choice",
    "field",
    "flag",
    "list_field",
    "mutable_field",
    "subparsers",
    "ArgumentParser",
    "InconsistentArgumentError",
]

from . import _version

__version__ = _version.get_versions()["version"]



import collections
from simple_parsing.helpers.hparams import uniform
from typing import Union
from simple_parsing.helpers.fields import field, subparsers
from simple_parsing.helpers.hparams.hyperparameters import HyperParameters
from dataclasses import dataclass
from .testutils import TestSetup
import functools

from simple_parsing import ArgumentParser


@dataclass
class A:
    foo: int = 123


@dataclass
class B:
    bar: float = 4.56


class TestWithDefault:
    @dataclass
    class Options(TestSetup):
        config: Union[A, B] = subparsers({"a": A, "b": B}, default=A(foo=0))

    def test_default_is_used_when_no_args_passed(self):
        assert self.Options.setup("").config == A(foo=0)

    def test_subparsers_work(self):
        assert self.Options.setup("a --foo 456").config == A(foo=456)
        assert self.Options.setup("b --bar 1.23").config == B(bar=1.23)


class TestWithDefaultFactory:
    @dataclass
    class Options(TestSetup):
        config: Union[A, B] = subparsers(
            {"a": A, "b": B}, default_factory=functools.partial(A, foo=0)
        )

    def test_default_is_used_when_no_args_passed(self):
        assert self.Options.setup("").config == A(foo=0)

    def test_subparsers_work(self):
        assert self.Options.setup("a --foo 456").config == A(foo=456)
        assert self.Options.setup("b --bar 1.23").config == B(bar=1.23)


class TestWithoutSubparsersField:
    @dataclass
    class Options(TestSetup):
        config: Union[A, B] = field(default_factory=functools.partial(A, foo=0))

    def test_default_is_used_when_no_args_passed(self):
        assert self.Options.setup("").config == A(foo=0)

    def test_subparsers_work(self):
        assert self.Options.setup("a --foo 456").config == A(foo=456)
        assert self.Options.setup("b --bar 1.23").config == B(bar=1.23)


class TestWithoutSubparsersField:
    @dataclass
    class Options(TestSetup):
        config: Union[A, B] = field(
            default=A(foo=0),
        )

    def test_default_is_used_when_no_args_passed(self):
        assert self.Options.setup("").config == A(foo=0)

    def test_subparsers_work(self):
        assert self.Options.setup("a --foo 456").config == A(foo=456)
        assert self.Options.setup("b --bar 1.23").config == B(bar=1.23)


def test_nesting_of_optional_subparsers():
    @dataclass
    class Bob:
        config: Union[A, B] = subparsers({"a": A, "b": B}, default=A(foo=0))

    @dataclass
    class Clarice:
        config: Union[A, B] = subparsers({"a": A, "b": B}, default=A(foo=0))

    @dataclass
    class NestedOptions(TestSetup):
        friend: Union[Bob, Clarice] = Bob()

    assert NestedOptions.setup("") == NestedOptions()
    assert NestedOptions.setup("bob") == NestedOptions(friend=Bob())
    assert NestedOptions.setup("bob a") == NestedOptions(friend=Bob(config=A()))
    assert NestedOptions.setup("bob a --foo 1") == NestedOptions(friend=Bob(config=A(foo=1)))
    assert NestedOptions.setup("bob b") == NestedOptions(friend=Bob(config=B()))
    assert NestedOptions.setup("bob b --bar 0.") == NestedOptions(friend=Bob(config=B(bar=0.0)))
    assert NestedOptions.setup("clarice") == NestedOptions(friend=Clarice())
    assert NestedOptions.setup("clarice a") == NestedOptions(friend=Clarice(config=A()))
    assert NestedOptions.setup("clarice a --foo 1") == NestedOptions(
        friend=Clarice(config=A(foo=1))
    )
    assert NestedOptions.setup("clarice b") == NestedOptions(friend=Clarice(config=B()))
    assert NestedOptions.setup("clarice b --bar 0.") == NestedOptions(
        friend=Clarice(config=B(bar=0.0))
    )


class ModelA(HyperParameters):
    foo: float = uniform(0, 1, default=0.5)


class ModelB(HyperParameters):
    bar: int = uniform(0, 10, default=5, discrete=True)


@dataclass
class Options(HyperParameters):
    model: Union[ModelA, ModelB] = field(default_factory=ModelA)


import pytest

import random


@pytest.mark.parametrize("seed", [123, 456, 789])
def test_sample_with_subparsers_field(seed: int):

    random.seed(seed)

    samples = [Options.sample() for _ in range(10)]
    assert not all(sample == Options() for sample in samples), samples

    model_types = [type(Options.sample().model) for _ in range(100)]
    assert len(set(model_types)) == 2
    assert 40 <= collections.Counter(model_types)[ModelA] <= 60



from dataclasses import dataclass, field, fields
from typing import List, Tuple

from simple_parsing import field

from .testutils import TestSetup


@dataclass
class Foo(TestSetup):
    a: int = 123

    b: str = "fooobar"
    c: Tuple[int, float] = (123, 4.56)

    d: List[bool] = field(default_factory=list)


@dataclass
class Bar(TestSetup):
    barry: Foo = field(default_factory=Foo)
    joe: "Foo" = field(default_factory=lambda: Foo(b="rrrrr"))
    z: "float" = 123.456
    some_list: "List[float]" = field(default_factory=[1.0, 2.0].copy)


def test_forward_ref():
    foo = Foo.setup()
    assert foo == Foo()

    foo = Foo.setup("--a 2 --b heyo --c 1 7.89")
    assert foo == Foo(a=2, b="heyo", c=(1, 7.89))


def test_forward_ref_nested():
    bar = Bar.setup()
    assert bar == Bar()
    assert bar.barry == Foo()
    bar = Bar.setup("--barry.a 2 --barry.b heyo --barry.c 1 7.89")
    assert bar.barry == Foo(a=2, b="heyo", c=(1, 7.89))
    assert isinstance(bar.joe, Foo)
